"""SDK client plugins."""

from .base import WithApiClient, PluginAccess, WaylayPlugin, WaylayTool, WaylayService

__all__ = [
    "WithApiClient",
    "PluginAccess",
    "WaylayPlugin",
    "WaylayTool",
    "WaylayService",
]



"""Auth exception utilities."""

import jose.exceptions as jwt_exc
from ..exceptions import WaylayError

_AUTH_MESSAGE_FOR_EXCEPTON_CLASS = [
    (jwt_exc.JWTClaimsError, "invalid token"),
    (jwt_exc.ExpiredSignatureError, "token expired"),
    (jwt_exc.JWTError, "invalid token"),
    (TypeError, "could not decode token"),
    (ValueError, "could not decode token"),
]


def _auth_message_for_exception(exception):
    for exc_class, msg in _AUTH_MESSAGE_FOR_EXCEPTON_CLASS:
        if isinstance(exception, exc_class):
            return msg
    return "could not decode token"


class AuthError(WaylayError):
    """Exception class for waylay authentication errors."""


class TokenParseError(AuthError):
    """Error parsing a waylay token."""

    def __init__(self, exc):
        """Create a token parse error."""
        super().__init__(_auth_message_for_exception(exc))



"""Client mixin for plugins."""

import sys
from typing import Type, Dict, Optional
from importlib.metadata import entry_points
import warnings

from .base import WaylayService, WaylayTool, WithApiClient, WaylayPlugin, PluginAccess
from ..api import ApiClient


class WithServicesAndTools(WithApiClient):
    """Client as loader of service and tool plugins."""

    services: PluginAccess[WaylayService]
    tools: PluginAccess[WaylayTool]
    _services: Dict[str, WaylayService]
    _tools: Dict[str, WaylayTool]

    def __init__(self, api_client: ApiClient):
        """Create a WaylayConfig instance."""
        super().__init__(api_client)
        self._services = {}
        self.services = PluginAccess[WaylayService](self._services, WaylayService)
        self._tools = {}
        self.tools = PluginAccess[WaylayTool](self._tools, WaylayTool)
        self._load_plugins()

    def _load_plugins(self):
        ep_group = "dynamic"
        ep_name = "waylay_sdk_plugins"
        if sys.version_info >= (3, 10):
            waylay_entry_points = entry_points(group=ep_group, name=ep_name)
        else:
            waylay_entry_points = [
                ep for ep in entry_points().get(ep_group, []) if ep.name == ep_name
            ]
        for ep in waylay_entry_points:
            for plugin_class in ep.load():
                self.register(plugin_class)

    def register(self, plugin_class: Type[WaylayPlugin]) -> Optional[WaylayPlugin]:
        """Register and instantiate plugin class."""
        if issubclass(plugin_class, WaylayService):
            service = plugin_class(self.api_client)
            self._services[service.name] = service
            return service
        if issubclass(plugin_class, WaylayTool):
            tool = plugin_class(
                self.api_client,
                services=self.services,
                tools=self.tools,
            )
            self._tools[tool.name] = tool
            return tool
        warnings.warn(f"Invalid plug class: {plugin_class}")
        return None

    def __getattr__(self, name: str):
        """Get plugin by name."""
        if name in self._services:
            return self._services[name]
        if name in self._tools:
            return self._tools[name]
        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{name}'"
        )



"""Default tools and services."""

from typing import Type, List

from .base import WaylayPlugin
from waylay.sdk.services.gateway import GatewayService

PLUGINS: List[Type[WaylayPlugin]] = [GatewayService]



"""REST client for the Waylay Platform."""

from __future__ import annotations
from typing import TYPE_CHECKING, Optional

from waylay.sdk.services.gateway import GatewayService

from .plugin.client import WithServicesAndTools
from .config.client import WaylayConfig, WithConfig, HttpClientOptions
from .api import ApiClient
from .api.http import AsyncClient


if TYPE_CHECKING:
    from .plugin.type_stubs import (
        AlarmsService,
        DataService,
        RegistryService,
        ResourcesService,
        RulesService,
        StorageService,
    )


class WaylayClient(WithConfig, WithServicesAndTools):
    """REST client for the Waylay Platform."""

    gateway: GatewayService
    alarms: "AlarmsService"
    data: "DataService"
    registry: "RegistryService"
    resources: "ResourcesService"
    rules: "RulesService"
    storage: "StorageService"

    def __init__(
        self,
        config: WaylayConfig | ApiClient,
        /,
        options: Optional[HttpClientOptions | AsyncClient] = None,
    ):
        """Create a WaylayConfig instance."""
        client_config: WaylayConfig
        api_client: ApiClient
        if isinstance(config, ApiClient):
            api_client = config
            client_config = api_client.config
            if options:
                api_client.set_options(options)
        else:
            client_config = config
            api_client = ApiClient(client_config, options)
        WithConfig.__init__(self, client_config)
        WithServicesAndTools.__init__(self, api_client)

    def __repr__(self):
        """Get a technical string representation of this instance."""
        return (
            f"<{self.__class__.__name__}("
            f"services=[{','.join(self._services.keys())}],"
            f"tools=[{','.join(self._tools.keys())}],"
            f"config={self.config}"
            ")>"
        )



"""API client."""

from __future__ import annotations
from typing import Mapping, Optional, Tuple, Union

from .._version import __version__
from ..config import WaylayConfig

from .exceptions import SyncCtxMgtNotSupportedError
from .serialization import WithSerializationSupport
from .http import AsyncClient, HttpClientOptions, Request, Response

RESTTimeout = Union[
    Optional[float],
    Tuple[Optional[float], Optional[float], Optional[float], Optional[float]],
]


class ApiClient(WithSerializationSupport):
    """Generic API client for OpenAPI client library builds.

    OpenAPI generic API client. This client handles the client-server
    communication, and is invariant across implementations. Specifics of
    the methods and models for each application are generated from the
    OpenAPI templates.

    :param configuration: configuration object for this client

    """

    config: WaylayConfig
    http_options: HttpClientOptions
    base_url: str
    _http_client: Optional[AsyncClient]

    def __init__(
        self,
        config: WaylayConfig,
        http_options: Optional[HttpClientOptions | AsyncClient] = None,
    ) -> None:
        """Create an instance."""
        self.config = config
        self.http_options = {}
        self._http_client = None
        self.base_url = self.config.gateway_url
        if http_options:
            self.set_options(http_options)

    def set_options(self, http_options: HttpClientOptions | AsyncClient | None):
        """Update http options on not open client."""
        if not self.is_closed:
            raise AttributeError("Cannot set options on open client.")
        if isinstance(http_options, AsyncClient):
            self._http_client = http_options
        elif isinstance(http_options, Mapping):
            self.http_options = http_options

    def clone(self, http_options: Optional[HttpClientOptions] = None):
        """Clone api client without http client."""
        http_options = (
            self.http_options
            if http_options is None
            else {**self.http_options, **http_options}
        )
        return self.__class__(self.config, http_options)

    @property
    def http_client(self) -> AsyncClient:
        """Get (or open) a http client."""
        if self._http_client is None or self.is_closed:
            self._http_client = self._init_http_client()
            return self._http_client
        return self._http_client

    @property
    def is_closed(self) -> bool:
        """Check that there is no active http client."""
        if not self._http_client:
            return True
        if self._http_client.is_closed:
            # someone else closed this http client, remove it.
            self._http_client = None
            return True
        return False

    async def __aenter__(self):
        """Initialize the http client."""
        if not self.is_closed:
            raise ValueError("Http client already in use.")
        self._http_client = self._init_http_client()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Close the http client."""
        await self.aclose()

    def __enter__(self):
        """Dissallow sync context management for now."""
        raise SyncCtxMgtNotSupportedError(self)

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Disallow sync context management."""
        raise SyncCtxMgtNotSupportedError(self)

    def _init_http_client(self) -> AsyncClient:
        opts = self.http_options
        client = AsyncClient(**opts)
        base_url = opts.get("base_url", self.config.gateway_url)
        client.base_url = base_url  # type: ignore
        auth = opts.get("auth", self.config.auth)
        client.auth = auth  # type: ignore
        client.headers.update({"User-Agent": f"waylay-sdk/python/{__version__}"})
        return client

    async def aclose(self):
        """Close the client."""
        if not self.is_closed:
            await self._http_client.aclose()
            self._http_client = None

    async def _request(self, *args, **kwargs) -> Response:
        """Invoke a http request."""
        return await self.http_client.request(*args, **kwargs)

    async def send(
        self, request: Request, *, stream: bool = False, **kwargs
    ) -> Response:
        """Send an http request."""
        return await self.http_client.send(request, stream=stream, **kwargs)



"""Exceptions."""

from typing import Any, Optional

from .http import Response

from ..exceptions import WaylayError, RequestError, RestResponseError


class ApiValueError(RequestError, ValueError):
    """Inappropriate argument value (of correct type) in a Waylay API request."""

    def __init__(self, msg, path_to_item=None) -> None:
        """Raise a value error.

        Args:
            msg (str): the exception message

        Keyword Args:
            path_to_item (list) the path to the exception in the
                received_data dict. None if unset

        """

        self.path_to_item = path_to_item
        full_msg = msg
        if path_to_item:
            full_msg = f"{msg} at {render_path(path_to_item)}"
        super(ApiValueError, self).__init__(full_msg)


class ApiError(RestResponseError):
    """Exception class wrapping the response data of a REST call."""

    data: Optional[Any]

    def __init__(
        self,
        response: Response,
        data: Optional[Any],
    ) -> None:
        """Create an instance."""
        super().__init__(response)
        self.data = data

    @classmethod
    def from_response(
        cls,
        response: Response,
        data: Optional[Any],
    ):
        """Create an instance from a REST exception response."""
        return cls(response, data)

    def __str__(self):
        """Get the string representation of the exception."""
        error_message = super().__str__()
        if self.data and self.data != self.response.content:
            error_message += f"\nResponse data: {self.data}"
        return error_message


class SyncCtxMgtNotSupportedError(WaylayError, TypeError):
    """Warn the user to use async context management."""

    def __init__(self, target):
        """Create the exception."""
        super().__init__(
            f"{target.__class__.__name__} object "
            "does not support the sync context management protocol.\n"
            'Use async context management instead: "await with ...:"'
        )


def render_path(path_to_item):
    """Return a string representation of a path."""
    result = ""
    for path in path_to_item:
        if isinstance(path, int):
            result += f"[{path}]"
        else:
            result += f"['{path}']"
    return result



"""Parse credentials."""

from typing import Dict, Any

from .model import (
    NoCredentials,
    ClientCredentials,
    WaylayCredentials,
    ApplicationCredentials,
    TokenCredentials,
)


def parse_credentials(json_obj: Dict[str, Any]) -> WaylayCredentials:
    """Convert a parsed json representation to a WaylayCredentials object."""
    cred_type = json_obj.get("type", None)
    if cred_type is None:
        raise ValueError("invalid json for credentials: missing type")

    for clz in [
        NoCredentials,
        ClientCredentials,
        ApplicationCredentials,
        TokenCredentials,
    ]:
        if clz.credentials_type == cred_type:  # type: ignore
            return clz(**{k: v for k, v in json_obj.items() if k != "type"})

    raise ValueError(f"cannot parse json for credential type {cred_type}")



"""Waylay SDK authentication."""

from .model import (
    WaylayCredentials,
    ClientCredentials,
    ApplicationCredentials,
    TokenCredentials,
    NoCredentials,
    WaylayToken,
)
from .provider import WaylayTokenAuth
from .parse import parse_credentials

from .exceptions import AuthError

__all__ = [
    "WaylayCredentials",
    "ClientCredentials",
    "ApplicationCredentials",
    "TokenCredentials",
    "NoCredentials",
    "WaylayToken",
    "WaylayTokenAuth",
    "parse_credentials",
    "AuthError",
]



"""Base exception class hierarchy for errors in the waylay client."""

from typing import Optional
from .api.http import Request, Response


class WaylayError(Exception):
    """Root class for all exceptions raised by this module."""


class ConfigError(WaylayError):
    """Exception class for waylay client configuration."""


class RequestError(WaylayError):
    """Exception class for request validation errors within the waylay client.

    Notifies errors in tools and utilities that are not directly related
    to a REST call.
    """


class RestConnectionError(RequestError):
    """Exception class for request errors that are caused by client connection errors."""


class RestError(WaylayError):
    """Exception class for failures to make a REST call."""


class RestRequestError(RestError):
    """Exception class for failures to prepare a REST call."""

    request: Optional[Request]

    def __init__(
        self,
        request: Optional[Request],
    ):
        """Create an instance."""
        self.request = request

    def __str__(self):
        """Get the string representation of the exception."""
        error_message = f"{self.__class__.__name__}"
        if self.request is None:
            return error_message
        req = self.request
        error_message += f"\nRequest: {req.method} {req.url}"
        if req.headers:
            error_message += f"\nRequest headers: {req.headers}"
        if hasattr(req, "_content"):
            error_message += f"\nRequest content: <bytes: len={len(req._content)}>"
        else:
            error_message += f"\nRequest content: <streaming: {req.stream}>"
        return error_message


class RestResponseError(RestRequestError):
    """Exception class wrapping the response data of a REST call."""

    response: Response

    def __init__(
        self,
        response: Response,
    ) -> None:
        """Create an instance."""
        super().__init__(response._request)
        self.response = response

    def __str__(self):
        """Get the string representation of the exception."""
        error_message = super().__str__()
        resp = self.response
        error_message += f"\nStatus: {resp.status_code}"
        error_message += f"\nReason: {resp.reason_phrase}"
        if resp.headers:
            error_message += f"\nResponse headers: {resp.headers}"
        if hasattr(self.response, "_content"):
            error_message += f"\nResponse content: <bytes: len={len(resp._content)}>"
        else:
            error_message += (
                f"\nResponse content: <streaming: len={resp.num_bytes_downloaded}>"
            )
        return error_message


class RestResponseParseError(RestResponseError):
    """Exception raised when a successfull http request (2XX) could not be parsed succesfully."""



"""Gateway Service."""

from __future__ import annotations

from typing import Any, Dict, Literal, TypeVar, overload
from pydantic import ConfigDict

from waylay.sdk.api.client import ApiClient
from waylay.sdk.api.http import HeaderTypes, QueryParamTypes, Response
from waylay.sdk.plugin.base import WaylayService, WithApiClient
from waylay.sdk.api._models import BaseModel as WaylayBaseModel, Model

T = TypeVar("T")


class GatewayService(WaylayService):
    """Gateway Service Class."""

    name = "gateway"
    title = "Gateway Service"

    about: AboutApi

    def __init__(self, api_client: ApiClient):
        """Create the gateway service."""
        super().__init__(api_client)
        self.about = AboutApi(api_client)


class AboutApi(WithApiClient):
    """About service methods."""

    @overload
    async def get(
        self,
        *,
        query: QueryParamTypes | None = None,
        raw_response: Literal[False] = False,
        select_path: Literal[""] = "",
        response_type: Literal[None] = None,
        headers: HeaderTypes | None = None,
        **kwargs,
    ) -> GatewayResponse: ...

    @overload
    async def get(
        self,
        *,
        query: QueryParamTypes | None = None,
        raw_response: Literal[False] = False,
        select_path: Literal[""] = "",
        response_type: T,
        headers: HeaderTypes | None = None,
        **kwargs,
    ) -> T: ...

    @overload
    async def get(
        self,
        *,
        query: QueryParamTypes | None = None,
        raw_response: Literal[True],
        select_path: Literal["_not_used_"] = "_not_used_",
        response_type: Literal[None] = None,  # not used
        headers: HeaderTypes | None = None,
        **kwargs,
    ) -> Response: ...

    @overload
    async def get(
        self,
        *,
        query: QueryParamTypes | None = None,
        raw_response: Literal[False] = False,
        select_path: str,
        response_type: Literal[None] = None,
        headers: HeaderTypes | None = None,
        **kwargs,
    ) -> Model: ...

    @overload
    async def get(
        self,
        *,
        query: QueryParamTypes | None = None,
        raw_response: Literal[False] = False,
        select_path: str,
        response_type: T,
        headers: HeaderTypes | None = None,
        **kwargs,
    ) -> T: ...

    async def get(
        self,
        *,
        query: QueryParamTypes | None = None,
        raw_response: bool = False,
        select_path: str = "",
        response_type: T | None = None,
        headers: HeaderTypes | None = None,
        **kwargs,
    ) -> GatewayResponse | T | Response | Model:
        """Get the status of the gateway.

        :param query: URL Query parameters.
        :type query: QueryParamTypes, optional
        :param raw_response: If true, return the http Response object instead of returning an api model object, or throwing an ApiError.
        :param select_path: Denotes the json path applied to the response object before returning it.
                Set it to the empty string `""` to receive the full response object.
        :param response_type: If specified, the response is parsed into an instance of the specified type.
        :param headers: Header parameters for this request
        :type headers: dict, optional
        :param `**kwargs`: Additional parameters passed on to the http client.
            See below.
        :Keyword Arguments:
            * timeout: a single numeric timeout in seconds,
                or a tuple of _connect_, _read_, _write_ and _pool_ timeouts.
            * stream: if true, the response will be in streaming mode
            * cookies
            * extensions
            * auth
            * follow_redirects: bool

        :return: Returns the result object if the http request succeeded with status code '2XX'.
        :raises APIError: If the http request has a status code different from `2XX`. This
            object wraps both the http Response and any parsed data.
        """
        response_type_map: Dict[str, Any] = (
            {"2XX": response_type}
            if response_type is not None
            else {
                "200": GatewayResponse if not select_path else Model,
            }
        )
        return await self.api_client.request(
            method="GET",
            resource_path="/",
            path_params={},
            params=query,
            headers=headers,
            response_type=response_type_map,
            select_path=select_path,
            raw_response=raw_response,
            **kwargs,
        )


class GatewayResponse(WaylayBaseModel):
    """Gateway status response."""

    name: str
    version: str
    health: str

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
        extra="allow",
    )



"""Httpx Authentication provider."""

from typing import Optional, Callable, Generator

import httpx

from .model import (
    WaylayCredentials,
    WaylayToken,
    TokenString,
    TokenCredentials,
    NoCredentials,
    ApplicationCredentials,
    ClientCredentials,
)
from .exceptions import AuthError


CredentialsCallback = Callable[[Optional[str]], WaylayCredentials]


class WaylayTokenAuth(httpx.Auth):
    """Authentication flow with a waylay token.

    Will automatically refresh an expired token.

    """

    current_token: Optional[WaylayToken]
    credentials: WaylayCredentials

    def __init__(
        self,
        credentials: WaylayCredentials,
        initial_token: Optional[TokenString] = None,
        credentials_callback: Optional[CredentialsCallback] = None,
    ):
        """Create a Waylay Token authentication provider."""
        self.credentials = credentials
        self.current_token = None

        if isinstance(credentials, TokenCredentials):
            initial_token = initial_token or credentials.token

        if initial_token:
            try:
                self.current_token = self._create_and_validate_token(initial_token)
            except AuthError:
                pass

        self.credentials_callback = credentials_callback

    def auth_flow(
        self, request: httpx.Request
    ) -> Generator[httpx.Request, httpx.Response, None]:
        """Authenticate a http request.

        Implements the authentication callback for the http client.

        """
        token = self.assure_valid_token()
        request.headers["Authorization"] = f"Bearer {token}"
        yield request

    def assure_valid_token(self) -> WaylayToken:
        """Validate the current token and request a new one if invalid."""
        if self.current_token:
            # token exists and is valid
            return self.current_token

        self.current_token = self._create_and_validate_token(
            self._request_token_string()
        )
        return self.current_token

    def _create_and_validate_token(self, token: TokenString) -> WaylayToken:
        return WaylayToken(token).validate()

    def _request_token_string(self) -> TokenString:
        """Request a token."""
        if isinstance(self.credentials, NoCredentials):
            if self.credentials_callback is not None:
                # TODO: where is this used? Clients need to update their
                # definition of the callback to use gateway URL as argument.
                self.credentials = self.credentials_callback(
                    self.credentials.accounts_url or self.credentials.gateway_url
                )
            else:
                raise AuthError("No credentials or credentials_callback provided.")

        if isinstance(self.credentials, TokenCredentials):
            raise AuthError(
                f"cannot refresh expired token with credentials "
                f"of type '{self.credentials.credentials_type}'"
            )

        if isinstance(self.credentials, ApplicationCredentials):
            raise AuthError(
                f"credentials of type {self.credentials.credentials_type} are not supported yet"
            )

        if isinstance(self.credentials, ClientCredentials):
            return _request_token(self.credentials)

        raise AuthError(
            f"credentials of type {self.credentials.credentials_type} are not supported"
        )


def _request_token(credentials: ClientCredentials) -> str:
    token_url_prefix = (
        credentials.accounts_url or f"{credentials.gateway_url}/accounts/v1"
    )
    token_url = f"{token_url_prefix}/tokens?grant_type=client_credentials"
    token_req = {
        "clientId": credentials.api_key,
        "clientSecret": credentials.api_secret,
    }
    try:
        token_resp = httpx.post(url=token_url, json=token_req)
        if token_resp.status_code != 200:
            raise AuthError(
                f"could not obtain waylay token: {token_resp.content!r} [{token_resp.status_code}]"
            )
        token_resp_json = token_resp.json()
    except httpx.HTTPError as exc:
        raise AuthError(f"could not obtain waylay token: {exc}") from exc
    return token_resp_json["token"]



"""Base classes for Waylay SDK Plugins."""

from typing import Iterator, Optional, Generic, Type, TypeVar, Mapping
from ..api import ApiClient, HttpClientOptions
from ..api.exceptions import SyncCtxMgtNotSupportedError


P = TypeVar("P", bound="WaylayPlugin")
PI = TypeVar("PI", bound="WaylayPlugin")


class WithApiClient:
    """Base Waylay plugin class."""

    api_client: ApiClient

    def __init__(self, api_client: ApiClient):
        """Create a Waylay SDK plugin."""
        self.api_client = api_client

    async def __aenter__(self):
        """Initialize the api client."""
        self(None)
        self.api_client = await self.api_client.__aenter__()
        return self

    def __call__(self, http_options: Optional[HttpClientOptions] = None):
        """Initialize the http client."""
        if self.api_client.is_closed:
            self.api_client.set_options(http_options)
        else:
            # create a new api client.
            # leave the previous client for other services.
            self.api_client = self.api_client.clone(http_options)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Close the http client."""
        await self.api_client.aclose()

    def __enter__(self):
        """Dissallows sync context management."""
        raise SyncCtxMgtNotSupportedError(self)

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Disallow sync context management."""
        raise SyncCtxMgtNotSupportedError(self)


class WaylayPlugin(WithApiClient):
    """Plugin for the Waylay SDK client."""

    name: str = "__NO_NAME__"
    title: str = "__NO_TITLE__"
    description: Optional[str] = None

    def __str__(self):
        """Get a plugin descripton."""
        return f"{self.name}: {self.title}"

    def __repr__(self):
        """Get a plugin representation."""
        return f"<{self.__class__.__name__}({self.name})>"


class WaylayService(WaylayPlugin):
    """Base Waylay service class."""


class PluginAccess(Mapping[str, P], Generic[P]):
    """A lookup API for tools or services."""

    base_class: Type[P]

    def __init__(self, items: Mapping[str, P], base_class: Type[P]):
        """Create accessor for SDK plugins."""
        self._items = items
        self.base_class = base_class

    def __getitem__(self, __key: str) -> P:
        """Get an SDK plugin by key."""
        return self._items.__getitem__(__key)

    def __iter__(self) -> Iterator[str]:
        """Iterate SDK plugin keys."""
        return self._items.__iter__()

    def __len__(self) -> int:
        """Count registred SDK plugin."""
        return self._items.__len__()

    def iter(self, item_class: Type[PI], name: Optional[str] = None) -> Iterator[PI]:
        """Iterate over the plugins that satisfy the requirements."""
        if name:
            plug = self._items.get(name, None)
            if isinstance(plug, item_class):
                yield plug
            return
        for plug in self._items.values():
            if isinstance(plug, item_class):
                yield plug

    def select(self, item_class: Type[PI], name: Optional[str] = None) -> Optional[PI]:
        """Select the first SDK plugin that satifies the class and name requirements."""
        return next(self.iter(item_class, name), None)

    def require(self, item_class: Type[PI], name: Optional[str] = None) -> PI:
        """Get the SDK plugin for the given class or raise a ConfigError."""
        item = self.select(item_class, name)
        if item is None:
            info = f"{item_class.__name__} '{name}'" if name else item_class.__name__
            raise AttributeError(f"{info} is not available.")
        return item

    def __getattr__(self, name: str) -> P:
        """Get a plugin by name."""
        return self.require(self.base_class, name)

    def __repr__(self):
        """Get string representation."""
        return repr(self._items)


class WaylayTool(WaylayPlugin):
    """A tool extension for the waylay sdk."""

    _services: PluginAccess[WaylayService]
    _tools: "PluginAccess[WaylayTool]"

    def __init__(
        self,
        api_client: ApiClient,
        services: PluginAccess[WaylayService],
        tools: "PluginAccess[WaylayTool]",
    ):
        """Create a Waylay Tool."""
        super().__init__(api_client)
        self._services = services
        self._tools = tools



"""Client configuration."""

from .model import (
    WaylayConfig,
    TenantSettings,
    DEFAULT_PROFILE,
    SERVICE_KEY_GATEWAY,
    SERVICE_KEY_ACCOUNTS,
    SERVICE_KEY_API,
)

__all__ = [
    "WaylayConfig",
    "TenantSettings",
    "DEFAULT_PROFILE",
    "SERVICE_KEY_GATEWAY",
    "SERVICE_KEY_ACCOUNTS",
    "SERVICE_KEY_API",
]



"""Utilities to handle waylay authentication."""

from datetime import datetime
from enum import Enum
from dataclasses import dataclass
from typing import Optional, ClassVar, Dict, Any, List
import json
import abc

import base64
import binascii
from jose import jwt, JWTError

from .exceptions import TokenParseError, AuthError


class CredentialsType(str, Enum):
    """Supported Waylay Authentication Methods.

    Note that username/password authentication (as used in our IDP at
    https://login.waylay.io)
    is not (yet) supported.

    """

    CLIENT = "client_credentials"
    APPLICATION = "application_credentials"
    TOKEN = "token"
    CALLBACK = "interactive"

    def __str__(self):
        """Get the string representation."""
        return self.value


TokenString = str


class WaylayCredentials(abc.ABC):
    """Base class for the representation of credentials to the waylay platform."""

    gateway_url: Optional[str] = None
    credentials_type: ClassVar[CredentialsType] = CredentialsType.CALLBACK

    # legacy
    accounts_url: Optional[str] = None

    @abc.abstractmethod
    def to_dict(self, obfuscate=True) -> Dict[str, Any]:
        """Convert the credentials to a json-serialisable representation."""

    @abc.abstractproperty
    def id(self) -> Optional[str]:
        """Get the main identifier for this credential."""
        return None

    def __repr__(self):
        """Show the implementing class and public information."""
        return f"<{self.__class__.__name__}({str(self)})>"

    def __str__(self):
        """Show the credential attributes, with secrets obfuscated."""
        return json.dumps(self.to_dict(obfuscate=True))

    @abc.abstractmethod
    def is_well_formed(self) -> bool:
        """Validate that these credentials are well-formed.

        This does not assure that they will lead to a succesfull
        authentication.

        """


@dataclass(repr=False)
class AccountsUrlMixin:
    """Dataclass mixin for the 'gateway_url' (legacy 'accounts_url') property."""

    gateway_url: Optional[str] = None
    accounts_url: Optional[str] = None


@dataclass(repr=False, init=False)
class ApiKeySecretMixin(AccountsUrlMixin):
    """Dataclass mixin for the 'api_key' and 'api_secret'."""

    api_key: str = ""
    api_secret: str = ""

    def __init__(
        self,
        api_key: str,
        api_secret: str,
        *,
        gateway_url: Optional[str] = None,
        accounts_url: Optional[str] = None,
    ):
        """Initialise with the api_key and api_secret."""
        super().__init__(gateway_url=gateway_url, accounts_url=accounts_url)
        self.api_key = api_key
        self.api_secret = api_secret

    @property
    def id(self) -> Optional[str]:
        """Get the main identifier for this credential."""
        return self.api_key

    @classmethod
    def create(
        cls,
        api_key: str,
        api_secret: str,
        *,
        gateway_url: Optional[str] = None,
        accounts_url: Optional[str] = None,
    ):
        """Create a client credentials object."""
        return cls(
            api_key=api_key,
            api_secret=api_secret,
            gateway_url=gateway_url,
            accounts_url=accounts_url,
        )

    def to_dict(self, obfuscate=True):
        """Convert the credentials to a json-serialisable representation."""
        return dict(
            type=self.credentials_type.value,
            api_key=self.api_key,
            api_secret="********" if obfuscate else self.api_secret,
            gateway_url=self.gateway_url,
            accounts_url=self.accounts_url,
        )

    def is_well_formed(self) -> bool:
        """Validate that these credentials are well-formed.

        This does not assure that they will lead to a succesfull
        authentication.

        """
        if not (self.api_key and self.api_secret):
            return False
        # api key are 12 bytes hexc encoded
        try:
            if len(bytes.fromhex(self.api_key)) != 12:
                return False
        except ValueError:
            return False
        # api secret are 24 bytes base64 encoded (rfc4648)
        try:
            if len(base64.b64decode(self.api_secret, validate=True)) != 24:
                return False
        except binascii.Error:
            return False
        return True


@dataclass(repr=False, init=False)
class NoCredentials(AccountsUrlMixin, WaylayCredentials):
    """Represents credentials which be asked via (interactive) callback when required."""

    credentials_type: ClassVar[CredentialsType] = CredentialsType.CALLBACK

    def to_dict(self, obfuscate=True):  # pylint: disable=unused-argument
        """Convert the credentials to a json-serialisable representation."""
        return dict(
            type=str(self.credentials_type),
            gateway_url=self.gateway_url,
            accounts_url=self.accounts_url,
        )

    def is_well_formed(self) -> bool:
        """Validate that these credentials are well-formed."""
        return True

    @property
    def id(self) -> Optional[str]:
        """Get the main identifier for this credential."""
        return None


@dataclass(repr=False, init=False)
class ClientCredentials(ApiKeySecretMixin, WaylayCredentials):
    """Waylay Credentials: api key and secret of type 'client_credentials'."""

    credentials_type: ClassVar[CredentialsType] = CredentialsType.CLIENT


@dataclass(repr=False, init=False)
class ApplicationCredentials(ApiKeySecretMixin, WaylayCredentials):
    """Waylay Credentials: api key and secret of type 'application_credentials'."""

    credentials_type: ClassVar[CredentialsType] = CredentialsType.APPLICATION
    tenant_id: str = ""


@dataclass(repr=False, init=False)
class TokenCredentials(AccountsUrlMixin, WaylayCredentials):
    """Waylay JWT Token credentials."""

    credentials_type: ClassVar[CredentialsType] = CredentialsType.TOKEN
    token: TokenString = ""

    def __init__(
        self,
        token: TokenString,
        *,
        gateway_url: Optional[str] = None,
        accounts_url: Optional[str] = None,
    ):
        """Create a TokenCredentials from a token string."""
        super().__init__(gateway_url=gateway_url, accounts_url=accounts_url)
        self.token = token

    @property
    def id(self) -> Optional[str]:
        """Get the main identifier for this credential."""
        try:
            token = WaylayToken(self.token)
            return f"[{token.domain}] {token.subject}"
        except AuthError as exc:
            return f"INVALID_TOKEN({exc})"

    def to_dict(self, obfuscate=True):
        """Get the credential attributes."""
        return dict(
            type=str(self.credentials_type),
            token="*********" if obfuscate else self.token,
            gateway_url=self.gateway_url,
            accounts_url=self.accounts_url,
        )

    def is_well_formed(self) -> bool:
        """Validate that these credentials are well-formed."""
        try:
            # WaylayToken constructor decodes the data without signature verification
            return WaylayToken(self.token).tenant is not None
        except TokenParseError:
            return False


class WaylayToken:
    """Holds a Waylay JWT token."""

    def __init__(self, token_string: str, token_data: Optional[Dict] = None):
        """Create a Waylay Token holder object from given token string or data."""
        self.token_string = token_string
        if token_data is None:
            try:
                token_data = jwt.decode(
                    token_string, "", options=dict(verify_signature=False)
                )
            except (TypeError, ValueError, JWTError) as exc:
                raise TokenParseError(exc) from exc
        self.token_data = token_data

    def validate(self) -> "WaylayToken":
        """Verify essential assertions, and its expiry state.

        This implementation does not verify the signature of a token, as
        this is seen the responsability of a server implementation.

        """
        if not self.token_string:
            raise AuthError("no token")

        if not self.token_data:
            raise AuthError("could not parse token data")

        if not self.tenant:
            raise AuthError("invalid token")

        # assert expiry
        if self.is_expired:
            raise AuthError("token expired")
        return self

    @property
    def tenant(self) -> Optional[str]:
        """Get the tenant id asserted by the token."""
        return self.token_data.get("tenant", None)

    @property
    def domain(self) -> Optional[str]:
        """Get the waylay domain asserted by the token."""
        return self.token_data.get("domain", None)

    @property
    def subject(self) -> Optional[str]:
        """Get the subject asserted by the token."""
        return self.token_data.get("sub", None)

    @property
    def licenses(self) -> List[str]:
        """Get the licenses asserted by the token."""
        return self.token_data.get("licenses", [])

    @property
    def groups(self) -> List[str]:
        """Get the groups asserted by the token."""
        return self.token_data.get("groups", [])

    @property
    def permissions(self) -> List[str]:
        """Get the permissions asserted by the token."""
        return self.token_data.get("permissions", [])

    @property
    def expires_at(self) -> Optional[datetime]:
        """Get the token expiry timestamp."""
        exp = self.token_data.get("exp", None)
        return None if exp is None else datetime.fromtimestamp(exp)

    @property
    def issued_at(self) -> Optional[datetime]:
        """Get the token issuance timestamp."""
        iat = self.token_data.get("iat", None)
        return None if iat is None else datetime.fromtimestamp(iat)

    @property
    def expires_seconds(self) -> int:
        """Get seconds until expiry."""
        exp = self.token_data.get("exp", None)
        return 0 if exp is None else exp - datetime.now().timestamp()

    @property
    def age(self) -> int:
        """Get seconds sinds issuance."""
        iat = self.token_data.get("iat", 0)
        return int(datetime.now().timestamp() - iat)

    @property
    def is_expired(self) -> bool:
        """Get the expiration state.

        True if a (previously valid) the token has expired.

        """
        if not isinstance(self.token_data, dict):
            return True
        exp = self.expires_at
        return exp is None or exp < datetime.now()

    @property
    def is_valid(self) -> bool:
        """Get the token validity.

        True if essential token data is present and is not expired.

        """
        return (
            self.tenant is not None
            and self.subject is not None
            and self.domain is not None
            and not self.is_expired
        )

    def to_dict(self):
        """Get the main token attributes."""
        return dict(
            tenant=self.tenant,
            domain=self.domain,
            subject=self.subject,
            expires_at=str(self.expires_at),
            is_valid=self.is_valid,
        )

    def __repr__(self) -> str:
        """Show the implementing class an main attributes."""
        return f"<{self.__class__.__name__}({json.dumps(self.to_dict())})>"

    def __str__(self) -> str:
        """Render the token string."""
        return self.token_string

    def __bool__(self) -> bool:
        """Get the validity of the token."""
        return self.is_valid



"""API client."""

from __future__ import annotations
import logging
import re
import datetime
from urllib.parse import quote
from inspect import isclass
from typing import (
    Any,
    Mapping,
    Optional,
    AsyncIterable,
    Type,
    Union,
    Iterable,
    Protocol,
    runtime_checkable,
)
from abc import abstractmethod
import warnings

from pydantic import BaseModel, ConfigDict, TypeAdapter, ValidationError
from pydantic_core import to_jsonable_python
from jsonpath_ng import parse as jsonpath_parse  # type: ignore[import-untyped]
from httpx import USE_CLIENT_DEFAULT
import httpx._client as httpxc

from .http import (
    AsyncClient,
    Request,
    Response,
    QueryParamTypes,
    HeaderTypes,
    RequestFiles,
    RequestContent,
    RequestData,
)
from .exceptions import ApiValueError, ApiError
from ._models import Model

_DEFAULT_RESPONSE_TYPE = Model
_PRIMITIVE_BYTE_TYPES = (bytes, bytearray)
_CLASS_MAPPING = {
    "date": datetime.date,
    "datetime": datetime.datetime,
    "Any": Any,
}
_ALLOWED_METHODS = ["GET", "HEAD", "DELETE", "POST", "PUT", "PATCH", "OPTIONS"]

_MODEL_TYPE_ADAPTER = TypeAdapter(Model)

log = logging.getLogger(__name__)

DEFAULT_SERIALIZATION_ARGS = {"by_alias": True, "exclude_none": True}


class WithSerializationSupport:
    """Serialization support for the SDK client."""

    base_url: str
    serialization_args = DEFAULT_SERIALIZATION_ARGS

    @property
    @abstractmethod
    def http_client(self) -> AsyncClient:
        """Get (or open) a http client."""

    async def request(
        self,
        method: str,
        resource_path: str,
        path_params: Optional[Mapping[str, str]] = None,
        *,
        params: Optional[Union[QueryParamTypes, Mapping, BaseModel]] = None,
        json: Optional[Any] = None,
        content: Optional[RequestContent] = None,
        files: Optional[RequestFiles] = None,
        data: Optional[RequestData] = None,
        headers: Optional[HeaderTypes] = None,
        cookies: httpxc.CookieTypes | None = None,
        timeout: httpxc.TimeoutTypes | None = None,
        extensions: httpxc.RequestExtensions | None = None,
        # Deserialization arguments
        response_type: Mapping[str, Type | None] | Type | None = None,
        select_path: str = "",
        raw_response: bool = False,
        # Additional parameters passed on to the http client
        **kwargs,
    ) -> Response | Any:
        """Perform a request with serialization and deserialization support."""

        # set aside send args
        send_args = {}
        for key in ["stream", "follow_redirects", "auth"]:
            if key in kwargs:
                send_args[key] = kwargs.pop(key)

        api_request = self.build_request(
            method,
            resource_path,
            path_params,
            params=params,
            json=json,
            content=content,
            files=files,
            data=data,
            headers=headers,
            cookies=cookies,
            timeout=timeout,
            extensions=extensions,
            **kwargs,
        )
        response = await self.http_client.send(api_request, **send_args)
        if raw_response:
            return response
        return self.deserialize(
            response,
            response_type=response_type,
            select_path=select_path,
            stream=send_args.get("stream", False),
        )

    def build_request(
        self,
        method: str,
        resource_path: str,
        path_params: Optional[Mapping[str, str]] = None,
        *,
        params: Optional[Union[QueryParamTypes, Mapping, BaseModel]] = None,
        json: Optional[Any] = None,
        content: Optional[RequestContent] = None,
        files: Optional[RequestFiles] = None,
        data: Optional[RequestData] = None,
        headers: Optional[HeaderTypes] = None,
        cookies: httpxc.CookieTypes | None = None,
        timeout: httpxc.TimeoutTypes | None = None,
        extensions: httpxc.RequestExtensions | None = None,
    ) -> Request:
        """Build the HTTP request params needed by the request."""
        method = _validate_method(method)
        url = _interpolate_resource_path(resource_path, path_params)
        return self.http_client.build_request(
            method,
            url,
            content=_convert_content(content),
            data=self.serialize(data),
            files=files,
            json=self.serialize(json),
            params=self.serialize(params),
            headers=headers,
            cookies=cookies,
            timeout=USE_CLIENT_DEFAULT if timeout is None else timeout,
            extensions=extensions,
        )

    def serialize(self, data):
        """Serialize to a jsonable python data structure."""
        return to_jsonable_python(data, **self.serialization_args)

    def deserialize(
        self,
        response: Response,
        *,
        response_type: Mapping[str, Type | None] | Type | None = None,
        select_path: str = "",
        stream: bool = False,
    ) -> Any:
        """Deserialize a http response into a python object.

        :param response_data: Response object to be deserialized.
        :param response_type: Response type to use for 2XX status codes,
            or a mapping per status code, including '2XX', '3XX', .. and
            '*' or 'default' wildcard keys.
        :param select_path: json path to be extracted from the json payload.
        :return: An instance of the type specified in the mapping.
        """
        if stream:
            warnings.warn(
                "Using `stream=True` is currently only supported with `raw_response=True`."
            )
            return response
        status_code = response.status_code
        response_type = _response_type_for_status_code(status_code, response_type)

        # deserialize response data
        return_data = None
        try:
            if response_type in _PRIMITIVE_BYTE_TYPES + tuple(
                t.__name__ for t in _PRIMITIVE_BYTE_TYPES
            ):
                return_data = response.content
            elif response_type is not None:
                try:
                    _data = response.json()
                    if select_path:
                        jsonpath_expr = jsonpath_parse(select_path)
                        match_values = [
                            match.value for match in jsonpath_expr.find(_data)
                        ]
                        _data = (
                            match_values[0]
                            if not re.search(r"\[(\*|.*:.*|.*,.*)\]", select_path)
                            else match_values
                        )
                except ValueError:
                    _data = response.text
                if _data is not None:
                    return_data = _deserialize(_data, response_type)
                else:
                    return_data = response.content
        finally:
            if not 200 <= status_code <= 299:
                raise ApiError.from_response(
                    response,
                    return_data,
                )
        return return_data


_CHUNK_SIZE = 65_536


@runtime_checkable
class Readable(Protocol):
    """Anything with a binary read method."""

    def read(self, n: int = -1) -> bytes:
        """Read a binary chunk"""


def _convert_content(content: Any):
    """Convert Iterable and Readable to AsyncIterable."""
    # do not handle cases httpx handles
    if isinstance(content, (bytes, str, AsyncIterable)):
        return content
    # non-dict Iterables fail when using async client, convert to async iterable.
    if isinstance(content, Iterable) and not isinstance(content, dict):

        async def _read_iterable_async():
            for chunk in content:
                yield chunk

        return _read_iterable_async()

    if isinstance(content, Readable):

        async def _read_reader_async():
            while chunk := content.read(_CHUNK_SIZE):
                yield chunk

        return _read_reader_async()
    return content


def _validate_method(method: str):
    method = method.upper()
    if method not in _ALLOWED_METHODS:
        raise ApiValueError(f"Method {method} is not supported.")
    return method


def _interpolate_resource_path(
    resource_path: str,
    path_params: Optional[Mapping[str, str]] = None,
):
    if not path_params:
        return resource_path
    for k, v in path_params.items():
        # specified safe chars, encode everything
        resource_path = resource_path.replace("{%s}" % k, quote(str(v)))
    return resource_path


def _deserialize(data: Any, klass: Any):
    """Deserializes response content into a `klass` instance."""
    if isinstance(klass, str) and klass in _CLASS_MAPPING:
        klass = _CLASS_MAPPING[klass]
    config = (
        ConfigDict(arbitrary_types_allowed=True)
        if not isclass(klass) or not issubclass(klass, BaseModel)
        else None
    )
    type_adapter = TypeAdapter(klass, config=config)
    try:
        return type_adapter.validate_python(data)
    except (TypeError, ValidationError) as exc:
        try:
            _deserialized = type_adapter.validate_python(
                data, strict=False, context={"skip_validation": True}
            )
            log.warning(
                "Failed to deserialize response into class %s, using backup non-validating deserializer instead.",
                klass,
                exc_info=exc,
                extra={"data": data, "class": klass},
            )
            return _deserialized
        except (TypeError, ValidationError):
            try:
                _deserialized = _MODEL_TYPE_ADAPTER.validate_python(data)
                log.warning(
                    "Failed to deserialize response into class %s, using backup generic model deserializer instead.",
                    klass,
                    exc_info=exc,
                    extra={"data": data, "class": klass},
                )
                return _deserialized
            except (TypeError, ValidationError) as exc2:
                log.warning(
                    "Failed to deserialize response as a generic Model, returning original data.",
                    exc_info=exc2,
                )
                return data


def _response_type_for_status_code(
    status_code,
    response_type: Mapping[str, Type | None] | Type | None,
):
    status_code_key = str(status_code)
    rt_map = (
        {"2XX": response_type} if not isinstance(response_type, dict) else response_type
    )
    if rt_map is None:
        return _DEFAULT_RESPONSE_TYPE
    for key in [status_code_key, f"{status_code_key[0]}XX", "default", "*"]:
        rt = rt_map.get(key)
        if rt is not None:
            return rt
    return _DEFAULT_RESPONSE_TYPE



from __future__ import annotations
from abc import ABC
from datetime import date, datetime
from decimal import Decimal
from inspect import isclass
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Union, get_type_hints

from typing_extensions import (
    Annotated,  # >=3.9
    Self,  # >=3.12
    TypeAliasType,  # >=3.12
)

if TYPE_CHECKING:
    from typing_extensions import TypeAlias  # >= Python 3.10

from pydantic import (
    BaseModel as PydanticBaseModel,
    ConfigDict,
    SerializationInfo,
    StrictStr,
    TypeAdapter,
    ValidationError,
    ValidationInfo,
    ValidatorFunctionWrapHandler,
    field_validator,
    model_serializer,
    model_validator,
)
from pydantic.functional_validators import ModelWrapValidatorHandler


class BaseModel(PydanticBaseModel, ABC):
    """Waylay base model class that adds some additional methods to Pydantic's `BaseModel`, including a custom validator and serializer."""

    @model_serializer(mode="wrap")
    def _model_serializer(
        self, handler: Callable, info: SerializationInfo
    ) -> Dict[StrictStr, Any]:
        """The default serializer of the model.

        * Excludes `None` fields that were not set at model initialization.
        """
        model_dict = handler(self, info)
        return {
            k: v
            for k, v in model_dict.items()
            if v is not None or k in self.model_fields_set
        }

    @model_validator(mode="wrap")  # type: ignore[arg-type]
    def _model_validator(
        cls, value: Any, handler: ModelWrapValidatorHandler, info: ValidationInfo
    ):
        """The default validator of the model.

        When validation is called with a `skip_validation=True` context (e.g. `cls.model_validate(data, context={"skip_validation": True})`), the model is constructed without validation.
        Any fields with a `Model` type will be constructed from their dict representation recursively.
        """
        context = info.context or {}
        try:
            return handler(value)
        except ValidationError:
            context = info.context or {}
            if context.get("skip_validation", False):
                model = cls.model_construct(**value)
                if not model.model_fields_set:
                    raise
                for field_name in model.model_fields_set:
                    field_value = getattr(model, field_name)
                    strict = (info.config or {}).get("strict")
                    try:
                        cls.__pydantic_validator__.validate_assignment(
                            model,
                            field_name,
                            field_value,
                            strict=strict,
                            context=context,
                        )
                    except BaseException:
                        pass
                return model
            else:
                raise

    @field_validator("*", mode="wrap")
    def _field_validator(
        cls, value: Any, handler: ValidatorFunctionWrapHandler, info: ValidationInfo
    ):
        """The default field validator of the model.

        When validation is called with a `skip_validation=True` context, the field is assigned without validation.
        If the field is a `Model` type, the model will be constructed from its dict representation recursively.
        """
        context = info.context or {}
        try:
            return handler(value)
        except ValidationError:
            context = info.context or {}
            if context.get("skip_validation", False):
                if info.field_name:
                    field_type = get_type_hints(cls).get(info.field_name, Any)
                    try:
                        config = (
                            ConfigDict(arbitrary_types_allowed=True, strict=False)
                            if not isclass(field_type)
                            or not issubclass(field_type, PydanticBaseModel)
                            else None
                        )
                        return TypeAdapter(field_type, config=config).validate_python(
                            value, strict=False, context=context
                        )
                    except ValidationError:
                        return value
            else:
                raise

    def to_dict(self) -> Dict[str, Any]:
        """Convert the model instance to dict."""
        return self.model_dump()

    def to_json(self) -> str:
        """Convert the model instance to a JSON-encoded string."""
        return self.model_dump_json()

    @classmethod
    def from_dict(cls, obj: dict) -> Self:
        """Create a model instance from a dict."""
        return cls.model_validate(obj)

    @classmethod
    def from_json(cls, json_data: str | bytes | bytearray) -> Self:
        """Create a model instance from a JSON-encoded string."""
        return cls.model_validate_json(json_data)


class _Model(BaseModel):
    """A simple model that allows all additional attributes."""

    model_config = ConfigDict(extra="allow")

    def __init__(self, /, **kwargs):
        super().__init__(**kwargs)
        for field in self.model_fields_set:  # pylint: disable=not-an-iterable
            setattr(
                self, field, _Model.__model_construct_recursive(getattr(self, field))
            )

    @classmethod
    def __model_construct_recursive(cls, obj: Any):
        if isinstance(obj, list):
            return [cls.__model_construct_recursive(inner) for inner in obj]
        elif isinstance(obj, dict):
            model = _Model(**obj)
            return model
        else:
            return obj


Primitive: TypeAlias = Union[
    str, bool, int, float, Decimal, bytes, datetime, date, object, None
]
Model: TypeAlias = TypeAliasType(  # type: ignore[valid-type]  #(https://github.com/python/mypy/issues/16614)
    "Model",
    Annotated[
        Union[List["Model"], "_Model", Primitive],  # type: ignore[misc,possible cyclic definition]
        "A basic model that acts like a `simpleNamespace`, or a collection over such models.",
    ],
)



"""Type imports for known services."""

from typing import TYPE_CHECKING
from .base import WaylayService


class _WaylayServiceStub(WaylayService):
    """Dummy Waylay service stub."""

    def __getattr__(self, name):
        """Get attribute."""
        raise ImportError(f"Waylay '{self.title}' service is not installed")

    def __bool__(self):
        """Cast to boolean."""
        return False


if TYPE_CHECKING:
    # registry service
    try:
        from waylay.services.registry.service import RegistryService  # type: ignore
    except ImportError:

        class _RegistryServiceStub(_WaylayServiceStub):
            name = "registry"
            title = "Function Registry (stub)"

        if not TYPE_CHECKING:
            RegistryService = _RegistryServiceStub

    # data broker service
    try:
        from waylay.services.data.service import DataService  # type: ignore
    except ImportError:

        class _DataServiceStub(_WaylayServiceStub):
            name = "data"
            title = "Data Broker (stub)"

        if not TYPE_CHECKING:
            DataService = _DataServiceStub

    # rules engine service
    try:
        from waylay.services.rules.service import RulesService  # type: ignore
    except ImportError:

        class _RulesServiceStub(_WaylayServiceStub):
            name = "rules"
            title = "Rules Engine (stub)"

        if not TYPE_CHECKING:
            RulesService = _RulesServiceStub

    # alarms service
    try:
        from waylay.services.alarms.service import AlarmsService  # type: ignore
    except ImportError:

        class _AlarmsServiceStub(_WaylayServiceStub):
            name = "alarms"
            title = "Alarms (stub)"

        if not TYPE_CHECKING:
            AlarmsService = _AlarmsServiceStub

    # resources service
    try:
        from waylay.services.resources.service import ResourcesService  # type: ignore
    except ImportError:

        class _ResourcesServiceStub(_WaylayServiceStub):
            name = "resources"
            title = "Resources (stub)"

        if not TYPE_CHECKING:
            ResourcesService = _ResourcesServiceStub

    # storage service
    try:
        from waylay.services.storage.service import StorageService  # type: ignore
    except ImportError:

        class _StorageServiceStub(_WaylayServiceStub):
            name = "storage"
            title = "Storage (stub)"

        if not TYPE_CHECKING:
            StorageService = _StorageServiceStub



"""Aliases for the http client."""

from __future__ import annotations
from typing import (
    TypedDict,
    Mapping,
    Optional,
    Callable,
    Any,
    Union,
    Sequence,
    Tuple,
    List,
    IO,
)
from typing_extensions import Required  # >=3.11

import httpx._client as httpx_types
import httpx

AsyncClient = httpx.AsyncClient
Response = httpx.Response
Request = httpx.Request

HeaderTypes = Union[
    Mapping[str, str],
    Mapping[bytes, bytes],
    Sequence[Tuple[str, str]],
    Sequence[Tuple[bytes, bytes]],
]

PrimitiveData = Optional[Union[str, int, float, bool]]
QueryParamTypes = Union[
    Mapping[str, Union[PrimitiveData, Sequence[PrimitiveData]]],
    List[Tuple[str, PrimitiveData]],
    Tuple[Tuple[str, PrimitiveData], ...],
    str,
    bytes,
]

RequestFiles = httpx_types.RequestFiles
RequestContent = Union[httpx_types.RequestContent, IO[bytes]]
RequestData = httpx_types.RequestData


class HttpRequestArguments(TypedDict, total=False):
    """Call parameters supported in an httpx request."""

    method: Required[str]
    url: Required[httpx_types.URLTypes]
    content: httpx_types.RequestContent
    data: httpx_types.RequestData
    files: RequestFiles
    json: Any
    params: QueryParamTypes
    headers: HeaderTypes
    cookies: httpx_types.CookieTypes
    auth: httpx_types.AuthTypes
    follow_redirects: bool
    timeout: httpx_types.TimeoutTypes
    extensions: httpx_types.RequestExtensions


class HttpClientOptions(TypedDict, total=False):
    """Options passed to the httpx client."""

    auth: Optional[httpx_types.AuthTypes]  # explicit None do disable auth
    params: httpx_types.QueryParamTypes
    headers: HeaderTypes
    cookies: httpx_types.CookieTypes
    verify: httpx_types.VerifyTypes
    cert: httpx_types.CertTypes
    http1: bool
    http2: bool
    proxy: httpx_types.ProxyTypes
    proxies: httpx_types.ProxiesTypes
    mounts: Mapping[str, Optional[httpx_types.AsyncBaseTransport]]
    timeout: httpx_types.TimeoutTypes
    follow_redirects: bool
    limits: httpx_types.Limits
    max_redirects: int
    event_hooks: Mapping[str, list[Callable[..., Any]]]
    base_url: str
    transport: httpx_types.AsyncBaseTransport
    app: Callable[..., Any]
    trust_env: bool
    default_encoding: str | Callable[[bytes], str]



"""Interactive authentication callback for client credentials."""

from typing import Optional
import logging
import re
import urllib.parse

from getpass import getpass
import httpx

from .model import WaylayCredentials, ClientCredentials, AuthError

DEFAULT_GATEWAY_URL = "https://api.waylay.io"
DEFAULT_ACCOUNTS_URL = "https://accounts-api.waylay.io"
ACCOUNTS_USERS_ME_PATH = "/accounts/v1/users/me"

_http = httpx
log = logging.getLogger(__name__)


def tell(message: str):
    """Show an interactive authentication message."""
    print(message)


def ask(prompt: str) -> str:
    """Prompt user for information."""
    return input(prompt)


def ask_secret(prompt: str) -> str:
    """Prompt user for credential information."""
    return getpass(prompt=prompt)


def ask_yes_no(prompt: str, default: Optional[bool] = None) -> bool:
    """Keep prompting the user until response starts with a true or false character."""
    while True:
        resp = ask(prompt)
        if not resp and default is not None:
            return default
        resp_char = resp.lower()[0]
        if resp_char in ("n", "f"):
            return False
        if resp_char in ("t", "y"):
            return True


def ask_gateway(default_gateway_url: str):
    """Ask for the gateway api in an interactive dialog."""
    gateway_url: str = _gateway_url_for(default_gateway_url)
    gateway_validated = False
    while not gateway_validated:
        tell(f"Proposed api gateway: {gateway_url}")
        tell("Please confirm, or specify a gateway by platform id, hostname or url.")
        tell(
            """Examples:
    'enterprise' (or 'api.waylay.io') for the Enterprise platform,
    'https://waylay-api.mycompany.com' as a custom endpoint url
        """
        )
        gateway_url = (
            ask(
                f"> Press enter to confirm, or specify an alternate gateway [{gateway_url}]: "
            )
            or gateway_url
        )
        gateway_url = _gateway_url_for(gateway_url)
        try:
            gateway_status_resp = _http.get(f"{gateway_url}{ACCOUNTS_USERS_ME_PATH}")
        except Exception as err:
            tell(f"Cannot connect to '{gateway_url}: {err}")
            gateway_url = _gateway_url_for(default_gateway_url)
        else:
            gateway_status = gateway_status_resp.status_code
            gateway_validated = gateway_status == 401
            if not gateway_validated:
                msg = (
                    "Should require authentication"
                    if gateway_status == 200
                    else gateway_status_resp.reason_phrase
                )
                tell(f"Not a gateway URL: '{gateway_url}': {msg}")
                gateway_url = _gateway_url_for(default_gateway_url)
            else:
                tell(f"Using gateway: {gateway_url}")
    return gateway_url


def request_migrate_to_gateway_interactive(profile, msg):
    """Asks to migrate to an api-gateway configuration."""
    tell(msg)
    tell(
        "NOTE: Migrating to an api-gateway will make this "
        "profile unusable for older waylay-py versions."
    )
    ans = ask_yes_no(f"Migrate '{profile}' to use an api-gateway? [Y/n]", True)
    if not ans:
        tell(f"Not migrating configuration profile '{profile}'")
    return ans


def request_client_credentials_interactive(
    default_gateway_url: str = DEFAULT_GATEWAY_URL,
) -> WaylayCredentials:
    """Asks interactively for client credentials.

    Default callback provider for an interactive WaylayConfig.

    """
    tell("Authenticating to the Waylay Platform")
    gateway_url = ask_gateway(default_gateway_url)
    tell("Please provide client credentials for the waylay data client.")
    credentials = ClientCredentials(api_key="", api_secret="", gateway_url=gateway_url)
    retry = 0
    while not credentials.is_well_formed() and retry < 3:
        api_key = ask(prompt="> apiKey : ").strip()
        api_secret = ask_secret(prompt="> apiSecret : ").strip()
        credentials = ClientCredentials(
            api_key=api_key, api_secret=api_secret, gateway_url=gateway_url
        )
        if not credentials.is_well_formed():
            retry += 1
            if retry >= 3:
                tell("Too many attempts, failing authentication")
                raise AuthError("Too many attempts, failing authentication")
            tell("Invalid apiKey or apiSecret, please retry")
    return credentials


def request_store_config_interactive(profile, save_callback):
    """Save interactively the storage of credentials as a profile."""
    if ask_yes_no(
        f"> Do you want to store these credentials with profile={profile}? [Y]: ", True
    ):
        save_location = save_callback()
        tell(
            f"Credential configuration stored as \n\t{save_location}\n"
            "Please make sure this file is treated securely.\n"
            "If compromised, _Revoke_ the api-key on the Waylay Console!"
        )


def _gateway_url_for(url_input: str):
    """Infer the gateway url from user input or existing accounts url."""
    url_input = (url_input or "").lower()
    if "accounts-api" in url_input:
        url_input = url_input.replace("accounts-api", "api")
    elif url_input in ("api", "", "enterprise"):
        url_input = "api.waylay.io"
    elif re.match(r"^[a-z0-9]+$", url_input):
        url_input = f"api-{url_input}.waylay.io"
    return _root_url_for(url_input)


def _root_url_for(host_or_url: str) -> str:
    scheme, loc, path, query, fragment = urllib.parse.urlsplit(host_or_url)

    if not scheme and not loc:
        # make sure any host name is converted in a https:// url
        scheme = "https"
        loc = path
        path = ""

    if path.endswith("/"):
        # tenant settings root urls are without trailing slash
        log.warning("Trailing slashes trimmed: %s", host_or_url)
        path = path.rstrip("/")

    return urllib.parse.urlunsplit([scheme, loc, path, query, fragment])



"""Config mixin for the SDK client."""

from __future__ import annotations
from typing import Optional

from ..auth import (
    WaylayCredentials,
    ClientCredentials,
    TokenCredentials,
)

from .model import (
    WaylayConfig,
    TenantSettings,
    DEFAULT_PROFILE,
    SERVICE_KEY_GATEWAY,
    SERVICE_KEY_ACCOUNTS,
)
from ..api import HttpClientOptions, AsyncClient


class WithConfig:
    """A configuration Mixin."""

    config: WaylayConfig
    http_options: HttpClientOptions

    def __init__(self, config: WaylayConfig, **args):
        """Create a configured entity."""
        self.config = config

    @classmethod
    def from_profile(
        cls,
        profile: str = DEFAULT_PROFILE,
        *,
        interactive=True,
        gateway_url=None,
        options: Optional[HttpClientOptions | AsyncClient] = None,
    ):
        """Create a WaylayClient named profile.

        Uses credentials from environment variables or a locally stored
        configuration.

        """
        return cls(
            WaylayConfig.load(
                profile, interactive=interactive, gateway_url=gateway_url
            ),
            options=options,
        )

    @classmethod
    def from_client_credentials(
        cls,
        api_key: str,
        api_secret: str,
        *,
        gateway_url=None,
        accounts_url=None,
        settings: Optional[TenantSettings] = None,
        options: Optional[HttpClientOptions | AsyncClient] = None,
    ):
        """Create a WaylayClient using the given client credentials."""
        credentials = ClientCredentials(
            api_key, api_secret, **_auth_urls(gateway_url, accounts_url, settings)
        )
        return cls.from_credentials(credentials, settings=settings, options=options)

    @classmethod
    def from_token(
        cls,
        token_string: str,
        *,
        gateway_url=None,
        accounts_url=None,
        settings: Optional[TenantSettings] = None,
        options: Optional[HttpClientOptions | AsyncClient] = None,
    ):
        """Create a WaylayClient using a waylay token."""
        credentials = TokenCredentials(
            token_string, **_auth_urls(gateway_url, accounts_url, settings)
        )
        return cls.from_credentials(credentials, settings=settings, options=options)

    @classmethod
    def from_credentials(
        cls,
        credentials: WaylayCredentials,
        settings: Optional[TenantSettings] = None,
        options: Optional[HttpClientOptions | AsyncClient] = None,
    ):
        """Create a WaylayClient using the given client credentials."""
        return cls(
            WaylayConfig(
                credentials,
                settings=settings,
            ),
            options=options,
        )


def _auth_urls(
    gateway_url=None, accounts_url=None, settings: Optional[TenantSettings] = None
):
    if settings:
        gateway_url = gateway_url or settings.get(SERVICE_KEY_GATEWAY)
        accounts_url = accounts_url or settings.get(SERVICE_KEY_ACCOUNTS)
    return {"gateway_url": gateway_url, "accounts_url": accounts_url}



"""The Waylay Python SDK."""

from .client import WaylayClient
from .config import WaylayConfig
from .auth import (
    WaylayCredentials,
    ClientCredentials,
    ApplicationCredentials,
    TokenCredentials,
    NoCredentials,
    WaylayToken,
)
from .plugin import WaylayService, WaylayTool, WaylayPlugin, PluginAccess
from .api import ApiClient
from .exceptions import WaylayError
from ._version import __version__

__all__ = [
    "WaylayClient",
    "WaylayConfig",
    "WaylayCredentials",
    "ClientCredentials",
    "ApplicationCredentials",
    "TokenCredentials",
    "NoCredentials",
    "WaylayToken",
    "WaylayService",
    "WaylayTool",
    "WaylayPlugin",
    "PluginAccess",
    "ApiClient",
    "WaylayError",
    "__version__",
]



"""Waylay Api Configuration."""

# export api classes
from .client import ApiClient, RESTTimeout
from .http import (
    AsyncClient,
    HttpClientOptions,
    HeaderTypes,
    QueryParamTypes,
    Request,
    Response,
    RequestFiles,
    RequestContent,
    RequestData,
)
from .exceptions import ApiError, ApiValueError
from ._models import Model, Primitive

__all__ = [
    "ApiClient",
    "RESTTimeout",
    "AsyncClient",
    "HttpClientOptions",
    "HeaderTypes",
    "QueryParamTypes",
    "Request",
    "Response",
    "RequestFiles",
    "RequestContent",
    "RequestData",
    "ApiError",
    "ApiValueError",
    "Model",
    "Primitive",
]
